""" Version of Topydo. """

VERSION = '0.16'

LICENSE = """Copyright (C) 2014-2017 Bram Schoenmakers
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law."""
